@@ -100,6 +100,19 @@ const struct value_string m3ua_ntfy_other_names[] = {
 	{ 0, NULL }
 };
 
+const char *m3ua_ntfy_info_name(uint16_t ntfy_status_type, uint16_t ntfy_status_info)
+{
+	switch (ntfy_status_type) {
+	case M3UA_NOTIFY_T_STATCHG:
+		return get_value_string(m3ua_ntfy_stchg_names, ntfy_status_info);
+	case M3UA_NOTIFY_T_OTHER:
+		return get_value_string(m3ua_ntfy_other_names, ntfy_status_info);
+	default:
+		break;
+	}
+	return "NULL";
+}
+
 static const struct value_string m3ua_iei_names[] = {
 	{ M3UA_IEI_INFO_STRING,		"INFO String" },
 	{ M3UA_IEI_ROUTE_CTX,		"Routing Context" },
@@ -473,7 +486,7 @@ int m3ua_decode_notify(struct osmo_xlm_prim_notify *npar, void *ctx,
  ***********************************************************************/
 
 /* Convert M3UA from xua_msg to msgb and set PPID/stream */
-static struct msgb *m3ua_to_msg(struct xua_msg *xua)
+struct msgb *m3ua_to_msg(struct xua_msg *xua)
 {
 	struct msgb *msg = xua_to_msg(M3UA_VERSION, xua);
 
@@ -493,26 +506,33 @@ static struct msgb *m3ua_to_msg(struct xua_msg *xua)
 	return msg;
 }
 
-/* transmit given xua_msg via given ASP */
+/* transmit given xua_msg via given ASP
+ * This function takes ownership of xua msg passed to it.
+ */
 static int m3ua_tx_xua_asp(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 {
-	struct msgb *msg = m3ua_to_msg(xua);
+	struct msgb *msg;
 
 	OSMO_ASSERT(asp->cfg.proto == OSMO_SS7_ASP_PROT_M3UA);
 
+	msg = m3ua_to_msg(xua);
+	xua_msg_free(xua);
 	if (!msg)
 		return -1;
 
+	/* msg becomes owned by osmo_ss7_asp_send here: */
 	return osmo_ss7_asp_send(asp, msg);
 }
 
 /*! \brief Send a given xUA message via a given M3UA Application Server
  *  \param[in] as Application Server through which to send \ref xua
  *  \param[in] xua xUA message to be sent
- *  \return 0 on success; negative on error */
+ *  \return 0 on success; negative on error
+ *
+ *  This function takes ownership of xua msg passed to it.
+ */
 int m3ua_tx_xua_as(struct osmo_ss7_as *as, struct xua_msg *xua)
 {
-	struct msgb *msg;
 	int rc;
 
 	OSMO_ASSERT(as->cfg.proto == OSMO_SS7_ASP_PROT_M3UA);
@@ -521,15 +541,11 @@ int m3ua_tx_xua_as(struct osmo_ss7_as *as, struct xua_msg *xua)
 	if (as->cfg.routing_key.context)
 		xua_msg_add_u32(xua, M3UA_IEI_ROUTE_CTX, as->cfg.routing_key.context);
 
-	msg = m3ua_to_msg(xua);
-	if (!msg)
-		return -1;
-
 	/* send the msg to the AS for transmission.  The AS FSM might
 	 * (depending on its state) enqueue it before transmission */
-	rc = osmo_fsm_inst_dispatch(as->fi, XUA_AS_E_TRANSFER_REQ, msg);
+	rc = osmo_fsm_inst_dispatch(as->fi, XUA_AS_E_TRANSFER_REQ, xua);
 	if (rc < 0)
-		msgb_free(msg);
+		xua_msg_free(xua);
 	return rc;
 }
 
@@ -555,6 +571,7 @@ struct m3ua_data_hdr *data_hdr_from_m3ua(struct xua_msg *xua)
 	return data_hdr;
 }
 
+/* This function takes ownership of xua msg passed to it. */
 static int m3ua_rx_xfer(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 {
 	struct xua_msg_part *na_ie = xua_msg_find_tag(xua, M3UA_IEI_NET_APPEAR);
@@ -570,7 +587,8 @@ static int m3ua_rx_xfer(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 			"%s(): unsupported message type: %s\n",
 			__func__,
 			get_value_string(m3ua_xfer_msgt_names, xua->hdr.msg_type));
-		return M3UA_ERR_UNSUPP_MSG_TYPE;
+		rc = M3UA_ERR_UNSUPP_MSG_TYPE;
+		goto ret_free;
 	}
 
 	/* Reject unsupported Network Appearance IE. */
@@ -581,13 +599,15 @@ static int m3ua_rx_xfer(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 			"Unsupported 'Network Appearance' IE '0x%08x' in message type '%s', sending 'Error'.\n",
 			na, get_value_string(m3ua_xfer_msgt_names, xua->hdr.msg_type));
 		if (na_ie->len != 4)
-			return M3UA_ERR_PARAM_FIELD_ERR;
-		return M3UA_ERR_INVAL_NET_APPEAR;
+			rc = M3UA_ERR_PARAM_FIELD_ERR;
+		else
+			rc = M3UA_ERR_INVAL_NET_APPEAR;
+		goto ret_free;
 	}
 
 	rc = xua_find_as_for_asp(&as, asp, rctx_ie);
 	if (rc)
-		return rc;
+		goto ret_free;
 
 	rate_ctr_inc2(as->ctrg, SS7_AS_CTR_RX_MSU_TOTAL);
 
@@ -602,6 +622,8 @@ static int m3ua_rx_xfer(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 		"%s(): M3UA data header: opc=%u=%s dpc=%u=%s\n",
 		__func__, xua->mtp.opc, osmo_ss7_pointcode_print(asp->inst, xua->mtp.opc),
 		xua->mtp.dpc, osmo_ss7_pointcode_print2(asp->inst, xua->mtp.dpc));
+	OSMO_ASSERT(xua->mtp.sls <= 0xf);
+	rate_ctr_inc2(as->ctrg, SS7_AS_CTR_RX_MSU_SLS_0 + xua->mtp.sls);
 
 	if (rctx_ie) {
 		/* remove ROUTE_CTX as in the routing case we want to add a new
@@ -609,8 +631,12 @@ static int m3ua_rx_xfer(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 		xua_msg_free_tag(xua, M3UA_IEI_ROUTE_CTX);
 	}
 
+	/* xua ownership is passed here: */
 	return m3ua_hmdc_rx_from_l2(asp->inst, xua);
-	/* xua will be freed by caller m3ua_rx_msg() */
+
+ret_free:
+	xua_msg_free(xua);
+	return rc;
 }
 
 static int m3ua_rx_mgmt_err(struct osmo_ss7_asp *asp, struct xua_msg *xua)
@@ -641,20 +667,7 @@ static int m3ua_rx_mgmt_ntfy(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 	m3ua_decode_notify(&ntfy, asp, xua);
 
 	type_name = get_value_string(m3ua_ntfy_type_names, ntfy.status_type);
-
-	switch (ntfy.status_type) {
-	case M3UA_NOTIFY_T_STATCHG:
-		info_name = get_value_string(m3ua_ntfy_stchg_names,
-						ntfy.status_info);
-		break;
-	case M3UA_NOTIFY_T_OTHER:
-		info_name = get_value_string(m3ua_ntfy_other_names,
-						ntfy.status_info);
-		break;
-	default:
-		info_name = "NULL";
-		break;
-	}
+	info_name = m3ua_ntfy_info_name(ntfy.status_type, ntfy.status_info);
 	LOGPASP(asp, DLM3UA, LOGL_NOTICE, "Received NOTIFY Type %s:%s (%s)\n",
 		type_name, info_name,
 		ntfy.info_string ? ntfy.info_string : "");
@@ -670,16 +683,24 @@ static int m3ua_rx_mgmt_ntfy(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 	return 0;
 }
 
+/* This function takes ownership of xua msg passed to it. */
 static int m3ua_rx_mgmt(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 {
+	int rc;
+
 	switch (xua->hdr.msg_type) {
 	case M3UA_MGMT_ERR:
-		return m3ua_rx_mgmt_err(asp, xua);
+		rc = m3ua_rx_mgmt_err(asp, xua);
+		break;
 	case M3UA_MGMT_NTFY:
-		return m3ua_rx_mgmt_ntfy(asp, xua);
+		rc = m3ua_rx_mgmt_ntfy(asp, xua);
+		break;
 	default:
-		return M3UA_ERR_UNSUPP_MSG_TYPE;
+		rc = M3UA_ERR_UNSUPP_MSG_TYPE;
 	}
+
+	xua_msg_free(xua);
+	return rc;
 }
 
 /* map from M3UA ASPSM/ASPTM to xua_asp_fsm event */
@@ -696,23 +717,26 @@ static const struct xua_msg_event_map m3ua_aspxm_map[] = {
 	{ M3UA_MSGC_ASPTM, M3UA_ASPTM_INACTIVE_ACK, XUA_ASP_E_ASPTM_ASPIA_ACK },
 };
 
-
+/* This function takes ownership of xua msg passed to it. */
 static int m3ua_rx_asp(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 {
 	int event;
+	int rc = 0;
 
 	/* map from the M3UA message class and message type to the XUA
 	 * ASP FSM event number */
 	event = xua_msg_event_map(xua, m3ua_aspxm_map,
 				  ARRAY_SIZE(m3ua_aspxm_map));
-	if (event < 0)
-		return M3UA_ERR_UNSUPP_MSG_TYPE;
+	if (event < 0) {
+		rc = M3UA_ERR_UNSUPP_MSG_TYPE;
+		goto ret_free;
+	}
 
 	/* deliver that event to the ASP FSM */
-	if (osmo_fsm_inst_dispatch(asp->fi, event, xua) < 0)
-		return M3UA_ERR_UNEXPECTED_MSG;
-
-	return 0;
+	rc = osmo_fsm_inst_dispatch(asp->fi, event, xua);
+ret_free:
+	xua_msg_free(xua);
+	return rc;
 }
 
 static int m3ua_rx_snm(struct osmo_ss7_asp *asp, struct xua_msg *xua);
@@ -740,9 +764,9 @@ int m3ua_rx_msg(struct osmo_ss7_asp *asp, struct msgb *msg)
 			"M3UA message\n");
 
 		if (hdr->version != M3UA_VERSION)
-			err = m3ua_gen_error_msg(M3UA_ERR_INVALID_VERSION, msg);
+			rc = M3UA_ERR_INVALID_VERSION;
 		else
-			err = m3ua_gen_error_msg(M3UA_ERR_PARAM_FIELD_ERR, msg);
+			rc = M3UA_ERR_PARAM_FIELD_ERR;
 		goto out;
 	}
 
@@ -750,7 +774,8 @@ int m3ua_rx_msg(struct osmo_ss7_asp *asp, struct msgb *msg)
 		xua_hdr_dump(xua, &xua_dialect_m3ua));
 
 	if (!xua_dialect_check_all_mand_ies(&xua_dialect_m3ua, xua)) {
-		err = m3ua_gen_error_msg(M3UA_ERR_MISSING_PARAM, msg);
+		rc = M3UA_ERR_MISSING_PARAM;
+		xua_msg_free(xua);
 		goto out;
 	}
 
@@ -762,7 +787,8 @@ int m3ua_rx_msg(struct osmo_ss7_asp *asp, struct msgb *msg)
 		/* The DATA message MUST NOT be sent on stream 0. */
 		if (msgb_sctp_stream(msg) == 0) {
 			rc = M3UA_ERR_INVAL_STREAM_ID;
-			break;
+			xua_msg_free(xua);
+			goto out;
 		}
 		rc = m3ua_rx_xfer(asp, xua);
 		break;
@@ -782,21 +808,16 @@ int m3ua_rx_msg(struct osmo_ss7_asp *asp, struct msgb *msg)
 	default:
 		LOGPASP(asp, DLM3UA, LOGL_NOTICE, "Received unknown M3UA "
 			"Message Class %u\n", xua->hdr.msg_class);
-		err = m3ua_gen_error_msg(M3UA_ERR_UNSUPP_MSG_CLASS, msg);
-		break;
+		rc = M3UA_ERR_UNSUPP_MSG_CLASS;
+		xua_msg_free(xua);
+		goto out;
 	}
 
-	if (rc > 0)
-		err = m3ua_gen_error_msg(rc, msg);
-
 out:
-	if (err) {
+	if (rc > 0)
+		err = m3ua_gen_error_msg(rc, msg);
+	if (err)
 		m3ua_tx_xua_asp(asp, err);
-		xua_msg_free(err);
-	}
-
-	xua_msg_free(xua);
-
 	return rc;
 }
 
@@ -896,7 +917,6 @@ void m3ua_tx_snm_available(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsig
 		xua = m3ua_encode_duna(rctx, num_rctx, aff_pc, num_aff_pc, info_string);
 
 	m3ua_tx_xua_asp(asp, xua);
-	xua_msg_free(xua);
 }
 
 /*! Transmit SSNM SCON message indicating congestion
@@ -930,7 +950,6 @@ void m3ua_tx_snm_congestion(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsi
 		xua_msg_add_data(xua, M3UA_IEI_INFO_STRING, strlen(info_string)+1, (const uint8_t *) info_string);
 
 	m3ua_tx_xua_asp(asp, xua);
-	xua_msg_free(xua);
 }
 
 /*! Transmit SSNM DUPU message indicating user unavailability.
@@ -946,10 +965,10 @@ void m3ua_tx_dupu(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int n
 {
 	struct xua_msg *xua = m3ua_encode_dupu(rctx, num_rctx, dpc, user, cause, info_str);
 	m3ua_tx_xua_asp(asp, xua);
-	xua_msg_free(xua);
 }
 
-/* received SNM message on ASP side */
+/* received SNM message on ASP side
+ * This function takes ownership of xua msg passed to it. */
 static int m3ua_rx_snm_asp(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 {
 	struct osmo_ss7_as *as = NULL;
@@ -958,7 +977,7 @@ static int m3ua_rx_snm_asp(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 
 	rc = xua_find_as_for_asp(&as, asp, rctx_ie);
 	if (rc)
-		return rc;
+		goto ret_free;
 
 	/* report those up the stack so both other ASPs and local SCCP users can be notified */
 	switch (xua->hdr.msg_type) {
@@ -988,14 +1007,16 @@ static int m3ua_rx_snm_asp(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 			xua_snm_rx_daud(asp, xua);
 		} else {
 			LOGPASP(asp, DLM3UA, LOGL_ERROR, "DAUD not permitted in ASP role\n");
-			return M3UA_ERR_UNSUPP_MSG_TYPE;
+			rc = M3UA_ERR_UNSUPP_MSG_TYPE;
 		}
 		break;
 	default:
-		return M3UA_ERR_UNSUPP_MSG_TYPE;
+		rc = M3UA_ERR_UNSUPP_MSG_TYPE;
 	}
 
-	return 0;
+ret_free:
+	xua_msg_free(xua);
+	return rc;
 }
 
 /* received SNM message on SG side */
